import os
import yaml
import colander
from colander import SchemaNode
from pprint import pprint
from datetime import datetime

from grano.core import db, celery
from grano.model import Schema, Attribute
from grano.model.schema import ENTITY_DEFAULT, RELATION_DEFAULT
from grano.logic.validation import Invalid, database_name
from grano.logic.references import ProjectRef, SchemaRef
from grano.plugins import notify_plugins
from grano.logic import attributes

TYPES_VALIDATOR = colander.OneOf(Attribute.DATATYPES.keys())
DEFAULTS = {
    'entity': ENTITY_DEFAULT,
    'relation': RELATION_DEFAULT
}


class AttributeValidator(colander.MappingSchema):
    name = SchemaNode(colander.String(),
                      validator=database_name)
    label = SchemaNode(colander.String(),
                       validator=colander.Length(min=3))
    description = SchemaNode(colander.String(),
                             missing='', default='')
    datatype = SchemaNode(colander.String(),
                          validator=TYPES_VALIDATOR,
                          missing='string')
    hidden = SchemaNode(colander.Boolean(),
                        missing=False)


class Attributes(colander.SequenceSchema):
    attribute = AttributeValidator()


class SchemaValidator(colander.MappingSchema):
    project = SchemaNode(ProjectRef())
    name = SchemaNode(colander.String(),
                      validator=database_name)
    label = SchemaNode(colander.String(),
                       validator=colander.Length(min=3))
    hidden = SchemaNode(colander.Boolean(),
                        missing=False)
    obj = SchemaNode(colander.String(),
                     validator=colander.OneOf(['entity', 'relation']))
    meta = SchemaNode(colander.Mapping(unknown='preserve'),
                      missing={})
    attributes = Attributes()


def validate(data):
    """ Validate the incoming data. """
    sane = SchemaValidator().deserialize(data)
    sref = SchemaRef(sane.get('project'))
    sane['parent'] = sref.decode(None, data.get('parent'))
    return sane


def save(data, schema=None):
    """ Create a schema. """
    data = validate(data)

    operation = 'create' if schema is None else 'update'
    if schema is None:
        schema = Schema()
        schema.name = data.get('name')
        schema.project = data.get('project')

    schema.label = data.get('label')
    schema.obj = data.get('obj')
    schema.hidden = data.get('hidden')
    schema.meta = data.get('meta')
    schema.parent = data.get('parent')

    if schema.name in DEFAULTS.values():
        schema.parent = None
    elif schema.parent is None or schema.is_circular():
        schema.parent = Schema.by_name(schema.project,
                                       DEFAULTS.get(schema.obj))

    schema.project.updated_at = datetime.utcnow()
    db.session.add(schema)

    inherited = [a.name for a in schema.inherited_attributes]
    names = []

    for attribute in data.get('attributes', []):
        if attribute.get('name') in inherited:
            continue
        attribute['schema'] = schema
        attr = attributes.save(attribute)
        schema.local_attributes.append(attr)
        names.append(attr.name)
    
    for attr in schema.local_attributes:
        if attr.name not in names:
            attributes.delete(attr)

    db.session.flush()
    _schema_changed(schema.project.slug, schema.name, operation)
    return schema


def delete(schema):
    if schema.parent is None and schema.children.count():
        return False
    _schema_changed(schema.project.slug, schema.name, 'delete')
    for child in schema.children:
        child.parent = schema.parent
    for attr in schema.local_attributes:
        attributes.delete(attr)
    db.session.delete(schema)
    return True


@celery.task
def _schema_changed(project_slug, schema_name, operation):
    """ Notify plugins about changes to a schema. """
    def _handle(obj):
        obj.schema_changed(project_slug, schema_name, operation)
    notify_plugins('grano.schema.change', _handle)


def import_schema(project, fh):
    data = yaml.load(fh.read())
    if isinstance(data, dict):
        data = [data]
    try:
        for cur in data:
            schema = Schema.by_name(project, cur.get('name'))
            cur['project'] = project
            save(cur, schema=schema)
        db.session.commit()
    except Invalid, inv:
        pprint(inv.asdict())


def export_schema(project, path):
    if not os.path.exists(path):
        os.makedirs(path)
    for schema in Schema.all().filter_by(project=project):
        fn = os.path.join(path, schema.name + '.yaml')
        with open(fn, 'w') as fh:
            dumped = yaml.safe_dump(schema.to_dict(schema),
                                    canonical=False,
                                    default_flow_style=False,
                                    indent=4)
            fh.write(dumped)



"""Add value_datetime_precision to Property

Revision ID: b64496669cd
Revises: 1b1b7e728dc4
Create Date: 2014-10-14 15:58:37.251624

"""

# revision identifiers, used by Alembic.
revision = 'b64496669cd'
down_revision = '1b1b7e728dc4'

from alembic import op
import sqlalchemy as sa


def upgrade():
    op.add_column(u'grano_property', sa.Column('value_datetime_precision', sa.Enum('year', 'month', 'day', 'hour', 'minute', 'second', 'microsecond', native_enum=False), nullable=True))


def downgrade():
    op.drop_column(u'grano_property', 'value_datetime_precision')



"""Add metadata to schema

Revision ID: 992aea84727
Revises: 383da0877c75
Create Date: 2014-07-09 14:27:28.657758

"""

# revision identifiers, used by Alembic.
revision = '992aea84727'
down_revision = '383da0877c75'

from alembic import op
import sqlalchemy as sa


def upgrade():
    op.drop_column('grano_schema', 'label_in')
    op.drop_column('grano_schema', 'label_out')
    op.add_column('grano_schema', sa.Column('meta', sa.Unicode(),
                                            nullable=True))


def downgrade():
    op.drop_column('grano_schema', 'meta')



from grano.core import db
from grano.model import Attribute, Property


def save(data):
    """ Create or update an attribute.
    CAUTION: This does not, on its own, validate any data."""

    schema = data.get('schema')
    name = data.get('name')
    obj = Attribute.by_schema_and_name(schema, name)
    if obj is None:
        obj = Attribute()
        obj.name = name
        obj.datatype = data.get('datatype')
        obj.schema = schema

    obj.label = data.get('label')
    obj.hidden = data.get('hidden')
    obj.inherited = data.get('inherited')
    obj.description = data.get('description')
    db.session.add(obj)
    return obj


def delete(attribute):
    """ Delete the attribute and all related properties. """
    q = db.session.query(Property)
    q = q.filter(Property.attribute == attribute)
    q.delete()
    db.session.delete(attribute)



from datetime import datetime

from sqlalchemy.orm import aliased

from grano.core import db
from grano.model.common import IntBase
from grano.model.attribute import Attribute

VALUE_COLUMNS = {
    'value_string': basestring,
    'value_datetime': datetime,
    'value_integer': int,
    'value_float': float,
    'value_boolean': bool
}
DATETIME_PRECISION = [
    'year',
    'month',
    'day',
    'time',
]


class Property(db.Model, IntBase):
    __tablename__ = 'grano_property'

    attribute_id = db.Column(db.Integer, db.ForeignKey('grano_attribute.id'))
    author_id = db.Column(db.Integer, db.ForeignKey('grano_account.id'))

    entity_id = db.Column(db.Unicode(), db.ForeignKey('grano_entity.id'),
                          index=True, nullable=True)
    relation_id = db.Column(db.Unicode(), db.ForeignKey('grano_relation.id'),
                            index=True, nullable=True)

    name = db.Column(db.Unicode(), index=True)

    value_string = db.Column(db.Unicode())
    value_integer = db.Column(db.Integer())
    value_float = db.Column(db.Float())
    value_datetime = db.Column(db.DateTime())
    value_datetime_precision = db.Column(db.Enum(*DATETIME_PRECISION, native_enum=False))
    value_boolean = db.Column(db.Boolean())
    value_file_id = db.Column(db.Integer(), db.ForeignKey('grano_file.id'))

    source_url = db.Column(db.Unicode())
    active = db.Column(db.Boolean())

    @property
    def value(self):
        # check file column first since file uses both
        # value_string and value_file_id
        if self.value_file_id is not None:
            return self.value_file_id
        for column in Attribute.DATATYPES.values():
            value = getattr(self, column)
            if value is not None:
                return value

    @classmethod
    def type_column(self, value):
        for name, typ in VALUE_COLUMNS.items():
            if isinstance(value, typ):
                return name
        return 'value_string'

    def to_dict_index(self):
        data = {
            'value': self.value,
            'source_url': self.source_url
        }
        if self.value_file_id is not None:
            data['file_url'] = self.value_string
        elif self.value_datetime is not None:
            data['value_precision'] = self.value_datetime_precision
        return data

    def to_dict_kv(self):
        return self.name, self.to_dict_index()

    def to_dict(self):
        name, data = self.to_dict_index()
        data['id'] = self.id
        data['name'] = name
        data['created_at'] = self.created_at
        data['updated_at'] = self.updated_at
        data['active'] = self.active
        return data


class PropertyBase(object):

    @property
    def active_properties(self):
        q = [p for p in self.properties if p.active]
        return q

    def __getitem__(self, name):
        for prop in self.active_properties:
            if prop.name == name:
                return prop

    def has_property(self, name):
        return self[name] is not None

    @classmethod
    def _filter_property(cls, q, name, value, only_active=True, alias=None):
        if alias is None:
            alias = cls
        Prop = aliased(Property)
        q = q.join(Prop, alias.properties)
        q = q.filter(Prop.name == name)
        column = getattr(Prop, Property.type_column(value))
        q = q.filter(column == value)
        if only_active:
            q = q.filter(Prop.active == True) # noqa
        return q



from flask import request
from restpager import Pager
from sqlalchemy.orm import aliased
from sqlalchemy.sql import func

from grano.lib.exc import BadRequest
from grano.lib.args import arg_bool
from grano.model import Project, Relation
from grano.model import Entity, Property, Schema, db
from grano.views import filters


def apply_facet_obj(q, facet_obj):
    q = q.add_entity(facet_obj)
    return q.group_by(facet_obj)


def apply_property_facet(q, facet, cls, parent_obj):
    """ Property facets are complicated because we don't want
    to facet over the whole property object, but merely it's
    value - which can be in one of any number of fields. """
    if '.' not in facet:
        raise BadRequest("Invalud facet: %s" % facet)
    _, name = facet.split('.', 1)
    facet_obj = aliased(cls)
    q = q.join(facet_obj, parent_obj.properties)
    q = q.filter(facet_obj.active == True)
    q = q.filter(facet_obj.name == name)
    columns = (facet_obj.value_string, facet_obj.value_integer,
               facet_obj.value_float, facet_obj.value_datetime,
               facet_obj.value_boolean)
    q = q.add_columns(*columns)
    return q.group_by(*columns)


def parse_entity_facets(entity_obj, full_facet, facet, q):
    """ Parse a facet related to a relation object and return a
    modified query. """
    # TODO: Status facet.
    # TODO: Author facet.
    if facet == 'project':
        facet_obj = aliased(Project)
        q = q.join(facet_obj, entity_obj.project)
        return apply_facet_obj(q, facet_obj)
    elif facet == 'schema':
        facet_obj = aliased(Schema)
        if not arg_bool('facet_%s_hidden' % full_facet, default=False):
            q = q.filter(facet_obj.hidden == False)
        q = q.join(facet_obj, entity_obj.schema)
        return apply_facet_obj(q, facet_obj)
    elif facet.startswith('properties.'):
        return apply_property_facet(q, facet, Property,
                                    entity_obj)
    elif facet.startswith('inbound.'):
        _, subfacet = facet.split('.', 1)
        rel_obj = aliased(Relation)
        q = q.join(rel_obj, entity_obj.inbound)
        return parse_relation_facets(rel_obj, full_facet, subfacet, q)
    elif facet.startswith('outbound.'):
        _, subfacet = facet.split('.', 1)
        rel_obj = aliased(Relation)
        q = q.join(rel_obj, entity_obj.outbound)
        return parse_relation_facets(rel_obj, full_facet, subfacet, q)
    else:
        raise BadRequest("Unknown facet: %s" % facet)


def parse_relation_facets(relation_obj, full_facet, facet, q):
    """ Parse a facet related to an entity and return a modified
    query. """
    if facet == 'project':
        facet_obj = aliased(Project)
        q = q.join(facet_obj, relation_obj.project)
        return apply_facet_obj(q, facet_obj)
    elif facet == 'schema':
        facet_obj = aliased(Schema)
        if not arg_bool('facet_%s_hidden' % full_facet, default=False):
            q = q.filter(facet_obj.hidden == False)
        q = q.join(facet_obj, relation_obj.schema)
        return apply_facet_obj(q, facet_obj)
    elif facet.startswith('properties.'):
        return apply_property_facet(q, facet, Property,
                                    relation_obj)
    elif facet.startswith('source.'):
        _, subfacet = facet.split('.', 1)
        ent_obj = aliased(Entity)
        q = q.join(ent_obj, relation_obj.source)
        return parse_entity_facets(ent_obj, full_facet, subfacet, q)
    elif facet.startswith('target.'):
        _, subfacet = facet.split('.', 1)
        ent_obj = aliased(Entity)
        q = q.join(ent_obj, relation_obj.target)
        return parse_entity_facets(ent_obj, full_facet, subfacet, q)
    else:
        raise BadRequest("Unknown facet: %s" % facet)


def results_process(q):
    for res in q:
        # pick the field that has a value, see the way properties
        # are queried. slight hack.
        count, value = res[0], max(res[1:])
        yield value, count


def for_entities():
    return make_facets(lambda: aliased(Entity),
                       filters.for_entities,
                       parse_entity_facets)


def for_relations():
    return make_facets(lambda: aliased(Relation),
                       filters.for_relations,
                       parse_relation_facets)


def make_facets(parent_alias, filter_func, parser_func):
    """ Return a set of facets based on the current query string. This
    will also consider filters set for the query, i.e. only show facets
    that match the current set of filters. """
    facets = {}
    for facet in request.args.getlist('facet'):
        parent_obj = parent_alias()
        q = db.session.query()
        facet_count = func.count(parent_obj.id)
        q = q.add_columns(facet_count)
        q = q.order_by(facet_count.desc())
        q = filter_func(q, parent_obj)
        q = parser_func(parent_obj, facet, facet, q)
        facets[facet] = Pager(q, name='facet_%s' % facet,
                              results_converter=results_process)
    return facets



"""empty message

Revision ID: 5a34d3b6a160
Revises: 3a3c75de71b
Create Date: 2014-03-14 15:51:15.830107

"""

# revision identifiers, used by Alembic.
revision = '5a34d3b6a160'
down_revision = '3a3c75de71b'

from alembic import op
import sqlalchemy as sa


def upgrade():
    ### commands auto generated by Alembic - please adjust! ###
    op.add_column('grano_property', sa.Column('value_boolean', sa.Boolean(), nullable=True))
    op.add_column('grano_property', sa.Column('value_datetime', sa.DateTime(), nullable=True))
    op.add_column('grano_property', sa.Column('value_float', sa.Float(), nullable=True))
    op.add_column('grano_property', sa.Column('value_integer', sa.Integer(), nullable=True))
    ### end Alembic commands ###


def downgrade():
    ### commands auto generated by Alembic - please adjust! ###
    op.drop_column('grano_property', 'value_integer')
    op.drop_column('grano_property', 'value_float')
    op.drop_column('grano_property', 'value_datetime')
    op.drop_column('grano_property', 'value_boolean')
    ### end Alembic commands ###



import os
from setuptools import setup, find_packages

README = os.path.join(os.path.dirname(__file__), 'README.rst')

REQUIREMENTS = os.path.join(os.path.dirname(__file__), 'requirements.txt')
REQUIREMENTS = open(REQUIREMENTS, 'r').read().splitlines()

VERSION = os.path.join(os.path.dirname(__file__), 'VERSION')
VERSION = open(VERSION, 'r').read().strip()

setup(
    name='grano',
    version=VERSION,
    description="An investigative toolkit for influence influence mapping",
    long_description=open(README, 'r').read(),
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Developers",
        "Operating System :: OS Independent",
        "Programming Language :: Python",
    ],
    keywords='sql graph sna networks journalism ddj entities',
    author='Code for Africa / ANCIR',
    author_email='support@codeforafrica.org',
    url='http://github.com/ANCIR/grano',
    license='MIT',
    packages=find_packages(exclude=['ez_setup', 'examples', 'tests']),
    namespace_packages=['grano'],
    package_data={'grano': ['fixtures/base.yaml']},
    include_package_data=True,
    zip_safe=False,
    install_requires=REQUIREMENTS,
    entry_points={
        'grano.entity.change': [],
        'grano.relation.change': [],
        'grano.project.change': [],
        'grano.schema.change': [],
        'grano.startup': [
            'bidi_create = grano.query.bidi:GenerateBidi',
            'levenshtein = grano.logic.reconcile:ConfigurePostgres'
        ],
        'grano.periodic': [
            'degrees = grano.logic.metrics:Degrees',
            'bidi_refresh = grano.query.bidi:GenerateBidi'
        ],
        'console_scripts': [
            'grano = grano.manage:run',
        ]
    },
    tests_require=[],
    test_suite='grano.test'
)



from flask import Blueprint, request
from restpager import Pager
from sqlalchemy.orm import aliased

from grano.lib.serialisation import jsonify
from grano.lib.args import object_or_404, request_data
from grano.model import Relation
from grano.logic import relations
from grano.logic.references import ProjectRef
from grano.views.cache import validate_cache
from grano.lib.exc import Gone
from grano.core import db
from grano.views import filters, facets, sorters
from grano import authz


blueprint = Blueprint('relations_api', __name__)


@blueprint.route('/api/1/relations', methods=['GET'])
def index():
    alias = aliased(Relation)
    q = db.session.query(alias)
    query = filters.for_relations(q, alias)
    query = sorters.for_entities(query, alias)
    pager = Pager(query)
    validate_cache(keys=pager.cache_keys())
    result = pager.to_dict()
    result['facets'] = facets.for_relations()
    return jsonify(result, index=True)


@blueprint.route('/api/1/relations', methods=['POST', 'PUT'])
def create():
    data = request_data({'author': request.account})
    project = ProjectRef().get(data.get('project'))
    data['project'] = project
    authz.require(authz.project_edit(project))
    relation = relations.save(data)
    db.session.commit()
    return jsonify(relation)


@blueprint.route('/api/1/relations/<id>', methods=['GET'])
def view(id):
    relation = object_or_404(Relation.by_id(id))
    authz.require(authz.relation_read(relation))
    return jsonify(relation)


@blueprint.route('/api/1/relations/<id>', methods=['POST', 'PUT'])
def update(id):
    relation = object_or_404(Relation.by_id(id))
    authz.require(authz.relation_edit(relation))
    data = request_data({'author': request.account})
    relation = relations.save(data, relation=relation)
    db.session.commit()
    return jsonify(relation)


@blueprint.route('/api/1/relations/<id>', methods=['DELETE'])
def delete(id):
    relation = object_or_404(Relation.by_id(id))
    authz.require(authz.relation_edit(relation))
    relations.delete(relation)
    db.session.commit()
    raise Gone()



import logging
from datetime import datetime

from grano.core import db
from grano.model import Pipeline, LogEntry
from grano.logic.accounts import console_account


def create(project, operation, label=None, config=None, author=None):
    """ Create a new pipeline object, in PENDING state. """
    pipeline = Pipeline()
    pipeline.status = Pipeline.STATUS_PENDING
    pipeline.project = project
    pipeline.label = label
    pipeline.operation = operation
    pipeline.percent_complete = 0
    pipeline.author = author or console_account()
    pipeline.config = config or {}
    db.session.add(pipeline)
    db.session.commit()
    return pipeline


def start(pipeline):
    """ Mark a given pipeline as running. """
    pipeline.status = Pipeline.STATUS_RUNNING
    pipeline.percent_complete = 0
    pipeline.started_at = datetime.utcnow()
    db.session.commit()


def finish(pipeline):
    """ Mark the given pipeline as complete or failed. """
    pipeline.percent_complete = 100
    if pipeline.has_errors():
        pipeline.status = Pipeline.STATUS_FAILED
    else:
        pipeline.status = Pipeline.STATUS_COMPLETE
    pipeline.ended_at = datetime.utcnow()
    db.session.commit()


def log(pipeline, level, message, error=None, details=None):
    """ Create a log entry related to the given pipeline. """
    entry = LogEntry()
    entry.pipeline = pipeline
    entry.level = level
    entry.message = message
    entry.error = error
    entry.details = details or {}
    db.session.add(entry)
    return entry


def log_debug(pipeline, message, error=None, details=None):
    return log(pipeline, logging.DEBUG, message, error=error, details=details)


def log_info(pipeline, message, error=None, details=None):
    return log(pipeline, logging.INFO, message, error=error, details=details)


def log_warn(pipeline, message, error=None, details=None):
    return log(pipeline, logging.WARNING, message, error=error,
               details=details)


def log_error(pipeline, message, error=None, details=None):
    return log(pipeline, logging.ERROR, message, error=error, details=details)



"""log entry models

Revision ID: 4d7168864daa
Revises: 32a25698e4bb
Create Date: 2014-04-06 21:42:28.317694

"""

# revision identifiers, used by Alembic.
revision = '4d7168864daa'
down_revision = '32a25698e4bb'

from alembic import op
import sqlalchemy as sa
from sqlalchemy.dialects import postgresql

def upgrade():
    op.create_table('grano_log_entry',
    sa.Column('created_at', sa.DateTime(), nullable=True),
    sa.Column('updated_at', sa.DateTime(), nullable=True),
    sa.Column('id', sa.Integer(), nullable=False),
    sa.Column('pipeline_id', sa.Integer(), nullable=True),
    sa.Column('level', sa.Unicode(), nullable=True),
    sa.Column('message', sa.Unicode(), nullable=True),
    sa.Column('error', sa.Unicode(), nullable=True),
    sa.Column('details', sa.Unicode(), nullable=True),
    sa.ForeignKeyConstraint(['pipeline_id'], ['grano_pipeline.id'], ),
    sa.PrimaryKeyConstraint('id')
    )


def downgrade():
    op.drop_table('grano_log_entry')



import logging

from grano.core import db
from grano.model import Entity, Relation, Project
from grano.logic import entities, relations, projects, accounts
from grano.logic.validation import Invalid


log = logging.getLogger(__name__)


class ObjectLoader(object):
    # Abstract parent

    def _setup(self, loader, schema):
        self.loader = loader
        self.schema = schema
        self.properties = {}
        self.update_criteria = set()

    def unique(self, name, only_active=True):
        """ Define a unique field for this entity or relation. Each unique
        key will be used to decide whether a record already exists and can
        be updated, or whether a new one must be created.

        :param name: The property name of the unique field.
        :param only_active: If set to ``False``, the check will include all
            historic values of the property as well as the current value.
        """
        self.update_criteria.add((name, only_active))

    def set(self, name, value, source_url=None):
        """ Set the value of a given property, optionally by attributing a
        source URL.

        :param name: The property name. This must be defined as part of one
            of the schemata that the entity or relation is associated with.
        :param value: The value to be set for this property. If it is
            ``None``, the property will not be set, but existing values of
            will be marked as inactive.
        :param source_url: A URL which will be set as the origin of this
            information.
        """
        source_url = source_url or self.source_url
        self.properties[name] = {
            'name': name,
            'value': value if value is None else unicode(value),
            'source_url': source_url,
            'active': True
            }


class EntityLoader(ObjectLoader):
    """ A factory object for entities, used to set the schemata and
    properties for an entity. """

    def __init__(self, loader, schema, source_url=None):
        self._setup(loader, schema)
        self.unique('name', only_active=False)
        self.source_url = source_url
        self._entity = None

    @property
    def entity(self):
        if self._entity is None:
            self.save()
        return self._entity

    def save(self):
        """ Save the entity to the database. Do this only once, after all
        properties have been set. """
        # fetch existing:
        q = Entity.all()
        q = q.filter(Entity.project == self.loader.project)
        for name, only_active in self.update_criteria:
            v = self.properties.get(name).get('value')
            q = Entity._filter_property(q, name, v, only_active=only_active)
        entity = q.first()

        try:
            data = {
                'project': self.loader.project,
                'author': self.loader.account,
                'schema': self.schema,
                'properties': self.properties
            }
            self._entity = entities.save(data, entity=entity)

        except Invalid, inv:
            if not self.loader.ignore_errors:
                raise
            log.warning("Validation error: %r", inv.asdict())


class RelationLoader(ObjectLoader):
    """ A factory object for relations, used to construct a relation by setting
    its schema, source entity, target entity and a set of properties. """

    def __init__(self, loader, schema, source, target, source_url=None):
        self._setup(loader, schema)
        self.source_url = source_url
        self.source = source
        self.target = target

    def save(self):
        """ Save the relation to the database. Do this only once, after all
        properties have been set. """

        # fetch existing:
        q = Relation.all()
        q = q.filter(Relation.project == self.loader.project)
        q = q.filter(Relation.source == self.source.entity)
        q = q.filter(Relation.target == self.target.entity)

        for name, only_active in self.update_criteria:
            v = self.properties.get(name).get('value')
            q = Entity._filter_property(q, name, v, only_active=only_active)
        relation = q.first()

        try:
            data = {
                'project': self.loader.project,
                'author': self.loader.account,
                'schema': self.schema,
                'properties': self.properties,
                'source': self.source.entity,
                'target': self.target.entity
            }
            self._relation = relations.save(data, relation=relation)
        except Invalid, inv:
            if not self.loader.ignore_errors:
                raise
            log.warning("Validation error: %r", inv.asdict())


class Loader(object):
    """ A loader is a factory object that can be used to make entities and
    relations in the database. It will perform some validation and handle
    database transactions. """

    def __init__(self, project_slug, source_url=None, project_label=None,
                 project_settings=None, account=None, ignore_errors=True):
        self.source_url = source_url
        self.account = account or accounts.console_account()
        self.ignore_errors = ignore_errors

        project = Project.by_slug(project_slug)
        project_settings = project_settings or \
            (project.settings if project else {})
        project_label = project_label or \
            (project.label if project else project_slug)
        self.project = projects.save({
            'slug': project_slug,
            'author': self.account,
            'label': project_label,
            'settings': project_settings
            }, project=project)

    def make_entity(self, schema, source_url=None):
        """ Create an entity loader, i.e. a construction helper for entities.

        :param schema: The schema that the entity should be associated with.
        :param source_url: A URL which will be made the default source for all
            properties defined on this entity.

        :returns: :py:class:`EntityLoader <grano.logic.loader.EntityLoader>`
        """
        source_url = source_url or self.source_url
        entity = EntityLoader(self, schema, source_url=source_url)
        return entity

    def make_relation(self, schema, source, target, source_url=None):
        """ Create a relation loader, i.e. a construction helper for relations.

        :param schema: A schema name for the relation.
        :param source: An :py:class:`EntityLoader
            <grano.logic.loader.EntityLoader>` which has been used to
            construct the source entity.
        :param target: A second :py:class:`EntityLoader
            <grano.logic.loader.EntityLoader>` which has been used to
            construct the target entity. Cannot be identical to the source.
        :param source_url: A URL which will be made the default source for all
            properties defined on this entity.

        :returns: :py:class:`RelationLoader
            <grano.logic.loader.RelationLoader>`
        """
        relation = RelationLoader(self, schema, source, target,
                                  source_url=source_url or self.source_url)
        return relation

    def persist(self):
        """ Save the created entiites and relations, i.e. commit the database
        transaction. """
        db.session.commit()



"""Add status field to Entity

Revision ID: 383da0877c75
Revises: 2372546f7078
Create Date: 2014-05-27 15:13:07.953084

"""

# revision identifiers, used by Alembic.
revision = '383da0877c75'
down_revision = '2372546f7078'

from alembic import op
import sqlalchemy as sa


def upgrade():
    op.add_column('grano_entity', sa.Column('status', sa.Integer))


def downgrade():
    op.drop_column('grano_entity', 'status')



import logging
import colander

from grano.core import db, celery
from grano.model import Relation
from grano.logic import properties as properties_logic
from grano.logic.references import ProjectRef, AccountRef
from grano.logic.references import SchemaRef, EntityRef
from grano.plugins import notify_plugins


log = logging.getLogger(__name__)


class RelationBaseValidator(colander.MappingSchema):
    author = colander.SchemaNode(AccountRef())
    project = colander.SchemaNode(ProjectRef())


def validate(data, relation):
    """ Due to some fairly weird interdependencies between the different elements
    of the model, validation of relations has to happen in three steps. """

    validator = RelationBaseValidator()
    sane = validator.deserialize(data)
    project = sane.get('project')

    schema_validator = colander.SchemaNode(colander.Mapping())
    schema_validator.add(colander.SchemaNode(SchemaRef(project),
                         name='schema'))
    schema_validator.add(colander.SchemaNode(EntityRef(project=project),
                         name='source'))
    schema_validator.add(colander.SchemaNode(EntityRef(project=project),
                         name='target'))

    sane.update(schema_validator.deserialize(data))

    sane['properties'] = properties_logic.validate('relation', relation,
                                                   project, sane.get('schema'),
                                                   data.get('properties', []))
    return sane


@celery.task
def _relation_changed(relation_id, operation):
    """ Notify plugins about changes to a relation. """
    def _handle(obj):
        obj.relation_changed(relation_id, operation)
    notify_plugins('grano.relation.change', _handle)


def save(data, relation=None):
    """ Save or update a relation with the given properties. """
    data = validate(data, relation)

    operation = 'create' if relation is None else 'update'
    if relation is None:
        relation = Relation()
        relation.project = data.get('project')
        relation.author = data.get('author')
        db.session.add(relation)

    relation.source = data.get('source')
    relation.target = data.get('target')
    relation.schema = data.get('schema')

    prop_names = set()
    for name, prop in data.get('properties').items():
        prop_names.add(name)
        prop['name'] = name
        prop['author'] = data.get('author')
        properties_logic.save(relation, prop)

    for prop in relation.properties:
        if prop.name not in prop_names:
            prop.active = False

    db.session.flush()
    _relation_changed.delay(relation.id, operation)
    return relation


def delete(relation):
    """ Delete the relation and its properties. """
    db.session.delete(relation)
    _relation_changed.delay(relation.id, 'delete')



from flask import Blueprint, request

from grano.lib.serialisation import jsonify
from grano.lib.args import object_or_404, request_data
from grano.model import Project
from grano.logic import imports
from grano import authz


blueprint = Blueprint('imports_api', __name__)


@blueprint.route('/api/1/projects/<slug>/_import', methods=['POST'])
def perform_import(slug):
    project = object_or_404(Project.by_slug(slug))
    authz.require(authz.project_edit(project))
    data = request_data({'project': project})
    pipeline = imports.make_importer(project, request.account, data)
    return jsonify(pipeline)



import unittest
import flask
from grano import authz
from grano.lib.args import single_arg
from grano.views import filters
from grano.core import db
from grano.model import Entity
from grano.test.test_authz import make_test_app, BaseAuthTestCase
from grano.test.test_authz import _project_and_permission
from werkzeug.datastructures import MultiDict
from werkzeug.exceptions import BadRequest


class AllEntitiesTestCase(BaseAuthTestCase):

    def setUp(self):
        self.app = make_test_app()
        Entity.all().delete()
        # Consistently include an extra private project with Entity
        # that should not show in any test results
        project, permission = _project_and_permission(private=True)
        entity = Entity(project=project, status=authz.PUBLISHED_THRESHOLD)
        db.session.add(entity)

    def test_all_entities__private(self):
        project, permission = _project_and_permission(private=True)
        entity = Entity(project=project, status=authz.PUBLISHED_THRESHOLD)
        db.session.add(entity)
        db.session.commit()
        with self.app.test_request_context():
            flask.session['id'] = 1
            self.app.preprocess_request()
            q = db.session.query(Entity)
            self.assertEqual(filters.for_entities(q, Entity).count(), 0)

    def test_all_entities__private_reader_published(self):
        project, permission = _project_and_permission(
            reader=True, private=True)
        entity = Entity(project=project, status=authz.PUBLISHED_THRESHOLD)
        db.session.add(entity)
        db.session.commit()
        with self.app.test_request_context():
            flask.session['id'] = 1
            self.app.preprocess_request()
            q = db.session.query(Entity)
            self.assertEqual(filters.for_entities(q, Entity).count(), 1)

    def test_all_entities__private_reader_draft(self):
        project, permission = _project_and_permission(
            reader=True, private=True)
        entity = Entity(project=project, status=authz.PUBLISHED_THRESHOLD-1)
        db.session.add(entity)
        db.session.commit()
        with self.app.test_request_context():
            flask.session['id'] = 1
            self.app.preprocess_request()
            q = db.session.query(Entity)
            self.assertEqual(filters.for_entities(q, Entity).count(), 0)

    def test_all_entities__private_editor_published(self):
        project, permission = _project_and_permission(
            editor=True, private=True)
        entity = Entity(project=project, status=authz.PUBLISHED_THRESHOLD)
        db.session.add(entity)
        db.session.commit()
        with self.app.test_request_context():
            flask.session['id'] = 1
            self.app.preprocess_request()
            q = db.session.query(Entity)
            self.assertEqual(filters.for_entities(q, Entity).count(), 1)

    def test_all_entities__private_editor_draft(self):
        project, permission = _project_and_permission(
            editor=True, private=True)
        entity = Entity(project=project, status=authz.PUBLISHED_THRESHOLD-1)
        db.session.add(entity)
        db.session.commit()
        with self.app.test_request_context():
            flask.session['id'] = 1
            self.app.preprocess_request()
            q = db.session.query(Entity)
            self.assertEqual(filters.for_entities(q, Entity).count(), 1)

    def test_all_entities__not_private_published(self):
        project, permission = _project_and_permission(private=False)
        entity = Entity(project=project, status=authz.PUBLISHED_THRESHOLD)
        db.session.add(entity)
        db.session.commit()
        with self.app.test_request_context():
            flask.session['id'] = 1
            self.app.preprocess_request()
            q = db.session.query(Entity)
            self.assertEqual(filters.for_entities(q, Entity).count(), 1)

    def test_all_entities__not_private_draft(self):
        project, permission = _project_and_permission(
            reader=True, private=False)
        entity = Entity(project=project, status=authz.PUBLISHED_THRESHOLD - 1)
        db.session.add(entity)
        db.session.commit()
        with self.app.test_request_context():
            flask.session['id'] = 1
            self.app.preprocess_request()
            q = db.session.query(Entity)
            self.assertEqual(filters.for_entities(q, Entity).count(), 0)


class SingleArgTestCase(BaseAuthTestCase):

    def setUp(self):
        self.app = make_test_app()

    def test_single_arg(self):
        with self.app.test_request_context('/?a=b'):
            self.assertEqual(single_arg('a'), 'b')

    def test_single_arg__bad_request(self):
        with self.app.test_request_context('/?a=b&a=c'):
            with self.assertRaises(BadRequest):
                single_arg('a')

    def test_single_arg__allow_empty_duplicates(self):
        with self.app.test_request_context('/?a=b&a='):
            self.assertEqual(single_arg('a'), 'b')




if __name__ == '__main__':
    unittest.main()



"""introduce project class

Revision ID: 40941ab93ba
Revises: 4f21a77e91be
Create Date: 2014-02-13 17:14:57.212052

"""

# revision identifiers, used by Alembic.
revision = '40941ab93ba'
down_revision = '4f21a77e91be'

from alembic import op
import sqlalchemy as sa


def upgrade():
    op.create_table('project',
    sa.Column('created_at', sa.DateTime(), nullable=True),
    sa.Column('updated_at', sa.DateTime(), nullable=True),
    sa.Column('id', sa.Integer(), nullable=False),
    sa.Column('slug', sa.Unicode(), nullable=True),
    sa.Column('label', sa.Unicode(), nullable=True),
    sa.Column('settings', sa.Unicode(), nullable=True),
    sa.Column('owner_id', sa.Integer(), nullable=True),
    sa.ForeignKeyConstraint(['owner_id'], ['account.id'], ),
    sa.PrimaryKeyConstraint('id')
    )


def downgrade():
    op.drop_table('project')
    



from datetime import datetime

import colander
from flask import url_for

from grano.core import db
from grano.logic import files as files_logic
from grano.model import Entity, Property, DATETIME_PRECISION


DATATYPE_TYPES = {
    'integer': colander.Integer(),
    'float': colander.Float(),
    'boolean': colander.Boolean(),
    'string': colander.String(),
    'datetime': colander.DateTime(default_tzinfo=None),
    'file': colander.Integer()
}


def validate_name(project, obj):
    """ Make sure that an entity's `name` attribute is unique
    within the scope of the project. """
    def check(name):
        entity = Entity.by_name(project, name)
        if entity is not None:
            if obj is None or obj.id != entity.id:
                return False
        return True
    name_unique = colander.Function(check,
                                    message="An entity with this name exists")
    return colander.All(name_unique, colander.Length(min=1))


def validate(obj_type, obj, project, schema, properties):
    """ Compile a validator for the given set of properties, based on
    the available schemata. """

    validator = colander.SchemaNode(colander.Mapping(), name='properties')
    for attr in schema.attributes:
        attrib = colander.SchemaNode(colander.Mapping(),
                                     name=attr.name,
                                     missing=colander.null)

        if attr.name == 'name' and obj_type == 'entity':
            attrib.add(colander.SchemaNode(colander.String(),
                       missing=colander.required, name='value',
                       validator=validate_name(project, obj)))
        else:
            T = DATATYPE_TYPES.get(attr.datatype)
            attrib.add(colander.SchemaNode(T, missing=None, name='value'))
            # add precision field to datetime attribute validator
            if attr.datatype == 'datetime':
                attrib.add(colander.SchemaNode(
                    colander.String(),
                    missing=None,
                    name='value_precision',
                    validator=colander.OneOf(DATETIME_PRECISION)
                ))

        attrib.add(colander.SchemaNode(colander.Boolean(),
                                       default=True, missing=True,
                                       name='active'))
        attrib.add(colander.SchemaNode(colander.String(),
                                       missing=None, name='source_url'))
        validator.add(attrib)

    properties = validator.deserialize(properties)
    out = {}
    for k, v in properties.items():
        if v != colander.null:
            out[k] = v
    return out


def save(obj, data, files=None):
    """ Set a property on the given object (entity or relation). This will
    either create a new property object or re-activate an existing object
    from the same source, if one exists. If the property is defined as
    ``active``, existing properties with the same name will be de-activated.

    WARNING: This does not, on its own, perform any validation.
    """
    prop = None

    for cand in obj.properties:
        if cand.name != data.get('name'):
            continue
        if cand.value == data.get('value'):
            prop = cand
        elif cand.active and data.get('active'):
            cand.active = False

    if prop is None:
        prop = Property()
        db.session.add(prop)
        if isinstance(obj, Entity):
            prop.entity = obj
        else:
            prop.relation = obj

    attribute = obj.schema.get_attribute(data.get('name'))
    if attribute.datatype == 'file':
        # if there is a file with the property name in `files`
        # a new file is created and the property updated
        file_key = data.get('name')
        if file_key in files:
            file_data = files.get(file_key)
            file = files_logic.save(data, file_data)
            prop.value_file_id = file.id
            prop.value_string = url_for('files_api.serve', id=file.id)
        elif data.get('value') is None:
            raise TypeError("File for property '%s' is required" % file_key)
    else:
        setattr(prop, attribute.value_column, data.get('value'))
        # set value of precision field for datetime property
        if attribute.datatype == 'datetime':
            prop.value_datetime_precision = data.get('value_precision')

    prop.name = data.get('name')
    prop.author = data.get('author')
    prop.attribute = attribute
    prop.active = data.get('active')
    prop.source_url = data.get('source_url')
    prop.updated_at = datetime.utcnow()
    obj.updated_at = datetime.utcnow()
    return prop


def delete(prop):
    """ Delete a property. """
    db.session.delete(prop)



"""add degree denormalizations

Revision ID: 38c7982f4160
Revises: 59d7b4f94cdf
Create Date: 2014-09-11 20:32:37.987989

"""

# revision identifiers, used by Alembic.
revision = '38c7982f4160'
down_revision = '59d7b4f94cdf'

from alembic import op
import sqlalchemy as sa


def upgrade():
    op.add_column(u'grano_entity', sa.Column('degree_in', sa.Integer(), nullable=True))
    op.add_column(u'grano_entity', sa.Column('degree_out', sa.Integer(), nullable=True))
    op.add_column(u'grano_entity', sa.Column('degree', sa.Integer(), nullable=True))


def downgrade():
    op.drop_column(u'grano_entity', 'degree_out')
    op.drop_column(u'grano_entity', 'degree_in')
    op.drop_column(u'grano_entity', 'degree')
    



from datetime import timedelta
import os

DEBUG = os.getenv("GRANO_DEBUG", True)
ASSETS_DEBUG = False
DEBUG_TIMING = False

# For production, it's essential that this is set:

SECRET_KEY = os.getenv("GRANO_SECRET_KEY", "test")
CACHE = True
CACHE_AGE = 84600

# Configure the database

SQLALCHEMY_DATABASE_URI = os.getenv("GRANO_DATABASE", "postgresql://grano:grano@postgres/grano")
APP_NAME = CELERY_APP_NAME = ES_INDEX = os.getenv("GRANO_APP_NAME", "grano")

# Force HTTPS here:
PREFERRED_URL_SCHEME = os.getenv("GRANO_URL_SCHEME", "http")

# You need to create an application on GitHub which can be used
# for OAuth sign-in.

GITHUB_CLIENT_ID = os.getenv("GRANO_GITHUB_CLIENT_ID", "da79a6b5868e690ab984")
GITHUB_CLIENT_SECRET = os.getenv("GRANO_GITHUB_CLIENT_SECRET", "1701d3bd20bbb29012592fd3a9c64b827e0682d6")

# TWITTER_API_KEY = 'UZYoBAfBzNluBlmBwPOGYw'
# TWITTER_API_SECRET = 'ngHaeaRPKA5BDQNXhPFmLWA1PvTA1kBGDaAJmc517E'

# FACEBOOK_APP_ID = '647877358607044'
# FACEBOOK_APP_SECRET = '5cb5c2181d0dc6976e97a55f90330165'

# Asynchronous task processing: For some tasks, grano will run
# some processing work in a delayed process, connected via a
# queue. If you do not wish to use delayed processing, you can
# turn it off with:

CELERY_ALWAYS_EAGER = True
CELERY_TASK_SERIALIZER = 'json'
CELERY_ACCEPT_CONTENT = ['json']

CELERYBEAT_SCHEDULE = {
    'periodic-tasks': {
        'task': 'grano.background.periodic',
        'schedule': timedelta(minutes=30)
    },
}

CELERY_TIMEZONE = 'UTC'

# Otherwise, just set the task broker URI which you intend to
# use:

CELERY_BROKER_URL = os.getenv("GRANO_CELERY_BROKER", "amqp://guest:guest@rabbitmq:5672")

DEFAULT_PLUGINS = ['degrees', 'bidi_create', 'bidi_refresh', 'levenshtein']
PLUGINS = ['ui']



import logging
import colander

from grano.core import db, celery
from grano.model import Entity
from grano.logic import properties as properties_logic
from grano.logic.references import ProjectRef, AccountRef
from grano.logic.references import SchemaRef, EntityRef
from grano.plugins import notify_plugins


log = logging.getLogger(__name__)


class EntityBaseValidator(colander.MappingSchema):
    author = colander.SchemaNode(AccountRef())
    project = colander.SchemaNode(ProjectRef())


class MergeValidator(colander.MappingSchema):
    orig = colander.SchemaNode(EntityRef())
    dest = colander.SchemaNode(EntityRef())


def validate(data, entity):
    """ Due to some fairly weird interdependencies between the different
    elements of the model, validation of entities has to happen in three
    steps. """
    validator = EntityBaseValidator()
    sane = validator.deserialize(data)
    project = sane.get('project')

    schema_validator = colander.SchemaNode(colander.Mapping())
    schema_validator.add(colander.SchemaNode(SchemaRef(project),
                         name='schema'))
    sane.update(schema_validator.deserialize(data))

    sane['properties'] = properties_logic.validate('entity', entity,
                                                   project, sane.get('schema'),
                                                   data.get('properties', []))
    return sane


@celery.task
def _entity_changed(entity_id, operation):
    """ Notify plugins about changes to an entity. """
    def _handle(obj):
        obj.entity_changed(entity_id, operation)
    notify_plugins('grano.entity.change', _handle)


def save(data, files=None, entity=None):
    """ Save or update an entity. """
    data = validate(data, entity)

    operation = 'create' if entity is None else 'update'
    if entity is None:
        entity = Entity()
        entity.project = data.get('project')
        entity.author = data.get('author')
        db.session.add(entity)

    entity.schema = data.get('schema')

    prop_names = set()
    for name, prop in data.get('properties').items():
        prop_names.add(name)
        prop['project'] = entity.project
        prop['name'] = name
        prop['author'] = data.get('author')
        properties_logic.save(entity, prop, files=files)

    for prop in entity.properties:
        if prop.name not in prop_names:
            prop.active = False

    db.session.flush()
    _entity_changed.delay(entity.id, operation)
    return entity


def delete(entity):
    """ Delete the entity and its properties, as well as any associated
    relations. """
    db.session.delete(entity)
    _entity_changed.delay(entity.id, 'delete')


def merge(source, dest):
    """ Copy all properties and relations from one entity onto another, then
    mark the source entity as an ID alias for the destionation entity. """
    if source.id == dest.id:
        return source

    if dest.same_as == source.id:
        return source

    if source.same_as == dest.id:
        return dest

    if dest.same_as is not None:
        # potential infinite recursion here.
        canonical = Entity.by_id(dest.same_as)
        if canonical is not None:
            return merge(source, canonical)

    if dest.schema.is_parent(source.schema):
        dest.schema = source.schema

    dest_valid = [a.name for a in dest.schema.attributes]
    dest_active = [p.name for p in dest.active_properties]
    for prop in source.properties:
        prop.entity = dest
        if prop.name in dest_active:
            prop.active = False
        if prop.name not in dest_valid:
            properties_logic.delete(prop)

    for rel in source.inbound:
        rel.target = dest
        db.session.add(rel)

    for rel in source.outbound:
        rel.source = dest
        db.session.add(rel)

    source.same_as = dest.id
    db.session.flush()
    _entity_changed.delay(dest.id, 'update')
    _entity_changed.delay(source.id, 'delete')
    return dest


def apply_alias(project, author, canonical_name, alias_name, source_url=None):
    """ Given two names, find out if there are existing entities for one or
    both of them. If so, merge them into a single entity - or, if only the
    entity associated with the alias exists - re-name the entity. """

    # Don't import meaningless aliases.
    if not len(canonical_name) or not len(alias_name):
        return log.info("Not an alias: %s", canonical_name)

    canonical = None

    # de-duplicate existing entities with the same name.
    known_names = set()
    for existing in Entity.by_name_many(project, canonical_name):

        for prop in existing.properties:
            if prop.name != 'name':
                continue
            known_names.add(prop.value)

            # make sure the canonical name is actually active
            if prop.value == canonical_name:
                prop.active = True
            else:
                prop.active = False

        if canonical is not None and canonical.id != existing.id:
            canonical = merge(existing, canonical)
        else:
            canonical = existing

    # Find aliases, i.e. entities with the alias name which are not
    # the canonical entity.
    q = Entity.by_name_many(project, alias_name)
    if canonical is not None:
        q = q.filter(Entity.id != canonical.id)
    aliases = q.all()

    # If there are no existing aliases with that name, add the alias
    # name to the canonical entity.
    if not len(aliases) and canonical is not None:
        if alias_name not in known_names:
            data = {
                'value': alias_name,
                'active': False,
                'name': 'name',
                'source_url': source_url
            }
            properties_logic.save(canonical, data)
            _entity_changed.delay(canonical.id, 'update')
        log.info("Alias: %s -> %s", alias_name, canonical_name)

    for alias in aliases:
        if canonical is None:
            # Rename an alias to its new, canonical name.
            data = {
                'value': canonical_name,
                'active': True,
                'name': 'name',
                'source_url': source_url
            }
            properties_logic.save(alias, data)
            _entity_changed.delay(alias.id, 'update')
            log.info("Renamed: %s -> %s", alias_name, canonical_name)
        else:
            # Merge two existing entities, declare one as "same_as"
            merge(alias, canonical)
            log.info("Mapped: %s -> %s", alias.id, canonical.id)

    db.session.commit()



import logging
import colander

from grano.core import db, celery
from grano.lib.data import CSVImporter
from grano.model import Pipeline, File
from grano.logic import pipelines, entities, loader
from grano.logic.references import FileRef, ProjectRef
from grano.logic.validation import Invalid


log = logging.getLogger(__name__)
MODES = ['aliases', 'entities', 'relations']


class ImportBaseValidator(colander.MappingSchema):
    file = colander.SchemaNode(FileRef())
    project = colander.SchemaNode(ProjectRef())
    source_url = colander.SchemaNode(colander.String(),
                                     empty=None, missing=None)
    mode = colander.SchemaNode(colander.String(),
                               validator=colander.OneOf(MODES))


def make_importer(project, account, data):
    """ Create an importer pipeline to represent the data
    import process which will be executed. """

    validator = ImportBaseValidator()
    sane = validator.deserialize(data)

    # TODO: validate mapping.
    config = {
        'mode': sane.get('mode'),
        'file': sane.get('file').id,
        'source_url': sane.get('source_url'),
        'mapping': data.get('mapping'),
        'relation_schema': data.get('relation_schema'),
        'source_schema': data.get('source_schema'),
        'target_schema': data.get('target_schema'),
        'entity_schema': data.get('entity_schema')
    }
    pipeline = pipelines.create(project, 'import',
                                sane.get('file').file_name, config, account)
    db.session.commit()
    run_importer.delay(pipeline.id)
    return pipeline


@celery.task
def run_importer(pipeline_id):
    """ Perform a raw data import with a given mode. """

    pipeline = Pipeline.by_id(pipeline_id)
    pipelines.start(pipeline)

    mode = pipeline.config.get('mode')
    file_id = pipeline.config.get('file')
    file_ = File.by_id(file_id)
    if file_ is None:
        pipeline.log_error(pipeline, 'File object deleted: %s' % file_id)
    elif mode == 'aliases':
        import_aliases(pipeline, file_.fh)
    else:
        import_objects(pipeline, file_.fh)

    pipelines.finish(pipeline)


def _row_source_url(pipeline, row):
    """ Determine the best available source URL for the given
    row of data. """
    for k, v in pipeline.config.get('mapping', {}).items():
        if v.get('attribute') == '_source_url':
            value = row.get(k, '').strip()
            if len(value):
                return value
    source_url = pipeline.config.get('source_url')
    if source_url is not None and len(source_url.strip()):
        return source_url
    return None


def import_aliases(pipeline, fh):
    """ Import aliases from a CSV source. This will not create
    new entities, but re-name existing entities or merge two
    entities if one's name is given as an alias for the other. """

    importer = CSVImporter(fh)
    canonical_column, alias_column = None, None
    for k, v in pipeline.config.get('mapping', {}).items():
        if v.get('attribute') == 'alias':
            alias_column = k
        elif v.get('attribute') == 'canonical':
            canonical_column = k

    for i, row in enumerate(importer):
        source_url = _row_source_url(pipeline, row)

        entities.apply_alias(pipeline.project, pipeline.author,
                             row.get(canonical_column),
                             row.get(alias_column),
                             source_url=source_url)

        if i % 100 == 0:
            percentage = int((float(i) / max(1, len(importer))) * 100)
            pipeline.percent_complete = percentage
            db.session.commit()


def import_objects(pipeline, fh):
    """ Import objects - either individual entities or relations
    and their involved entities (the target and source) - from a
    CSV file. """

    # Code is a bit ugly as this handles two cases at once:
    #  mode 'relations' where we import a source, target and relation
    #  mode 'entities' where we only import a single entity

    config = pipeline.config
    mode = config.get('mode')
    mapping = config.get('mapping')

    importer = CSVImporter(fh)
    loader_ = loader.Loader(pipeline.project.slug, account=pipeline.author,
                            ignore_errors=True)

    for i, row in enumerate(importer):
        try:
            url = _row_source_url(pipeline, row)
            rel_data = {}
            source = loader_.make_entity(config.get('source_schema'),
                                         source_url=url)
            target = loader_.make_entity(config.get('target_schema'),
                                         source_url=url)
            entity = loader_.make_entity(config.get('entity_schema'),
                                         source_url=url)

            # Try to assign each column to the appropriate object in this
            # loader.
            for column, spec in mapping.items():
                attr = spec.get('attribute')
                obj = spec.get('object')
                value = row.get(column)

                if not attr or not len(attr.strip()):
                    continue

                if mode == 'entities':
                    entity.set(attr, value)
                elif obj == 'relation':
                    rel_data[attr] = value
                elif obj == 'source':
                    source.set(attr, value)
                elif obj == 'target':
                    target.set(attr, value)

            # Relation can only be saved once the entities are available,
            # hence we're storing the relation property values and now
            # making the whole thing.
            if mode == 'relations':
                source.save()
                target.save()
                rel = loader_.make_relation(config.get('relation_schema'),
                                            source, target, source_url=url)
                for k, v in rel_data.items():
                    rel.set(k, v)
                rel.save()
            else:
                entity.save()

            # indicate progress, and commit every now and then.
            if i % 100 == 0:
                percentage = int((float(i) / max(1, len(importer))) * 100)
                pipeline.percent_complete = percentage
                loader_.persist()
        except Invalid, inv:
            pipelines.log_warn(pipeline, unicode(inv), 'Invalid data',
                               inv.as_dict())
        except Exception, exc:
            pipelines.log_error(pipeline, unicode(exc), 'Error', {})



from grano.core import db
from grano.model.common import IntBase


class Attribute(db.Model, IntBase):
    __tablename__ = 'grano_attribute'

    DATATYPES = {
        'string': 'value_string',
        'integer': 'value_integer',
        'float': 'value_float',
        'datetime': 'value_datetime',
        'boolean': 'value_boolean',
        'file': 'value_file_id'
    }

    name = db.Column(db.Unicode())
    label = db.Column(db.Unicode())
    description = db.Column(db.Unicode())
    hidden = db.Column(db.Boolean())
    datatype = db.Column(db.Unicode())

    schema_id = db.Column(db.Integer, db.ForeignKey('grano_schema.id'))
    properties = db.relationship('Property', backref='attribute',
                                 cascade='all, delete, delete-orphan',
                                 lazy='dynamic')

    @property
    def value_column(self):
        return self.DATATYPES.get(self.datatype)

    @classmethod
    def all_named(cls, name):
        q = db.session.query(cls)
        q = q.filter_by(name=name)
        return q.all()

    @classmethod
    def by_schema_and_name(cls, schema, name):
        q = db.session.query(cls)
        q = q.filter_by(schema=schema)
        q = q.filter_by(name=name)
        return q.first()

    def to_index(self):
        return {
            'name': self.name,
            'label': self.label,
            'datatype': self.datatype
        }

    def to_dict(self):
        data = self.to_index()
        data['id'] = self.id
        data['hidden'] = self.hidden
        if self.description and len(self.description):
            data['description'] = self.description
        return data



"""pipeline models

Revision ID: 32a25698e4bb
Revises: 52f4346513a8
Create Date: 2014-04-06 21:40:28.224450

"""

# revision identifiers, used by Alembic.
revision = '32a25698e4bb'
down_revision = '52f4346513a8'

from alembic import op
import sqlalchemy as sa
from sqlalchemy.dialects import postgresql

def upgrade():
    op.create_table('grano_pipeline',
    sa.Column('created_at', sa.DateTime(), nullable=True),
    sa.Column('updated_at', sa.DateTime(), nullable=True),
    sa.Column('id', sa.Integer(), nullable=False),
    sa.Column('operation', sa.Unicode(), nullable=True),
    sa.Column('status', sa.Unicode(), nullable=True),
    sa.Column('percent_complete', sa.Integer(), nullable=True),
    sa.Column('project_id', sa.Integer(), nullable=True),
    sa.Column('author_id', sa.Integer(), nullable=True),
    sa.Column('config', sa.Unicode(), nullable=True),
    sa.Column('started_at', sa.DateTime(), nullable=True),
    sa.Column('ended_at', sa.DateTime(), nullable=True),
    sa.ForeignKeyConstraint(['author_id'], ['grano_account.id'], ),
    sa.ForeignKeyConstraint(['project_id'], ['grano_project.id'], ),
    sa.PrimaryKeyConstraint('id')
    )


def downgrade():
    op.drop_table('grano_pipeline')
    



import logging

logging.basicConfig(level=logging.DEBUG)


# shut up useless SA warning:
import warnings
warnings.filterwarnings('ignore',
                        'Unicode type received non-unicode bind param value.')
from sqlalchemy.exc import SAWarning
warnings.filterwarnings('ignore', category=SAWarning)

# specific loggers
requests_log = logging.getLogger("requests")
requests_log.setLevel(logging.WARNING)

urllib3_log = logging.getLogger("urllib3")
urllib3_log.setLevel(logging.WARNING)

stevedore_log = logging.getLogger("stevedore")
stevedore_log.setLevel(logging.WARNING)



"""empty message

Revision ID: 3a3c75de71b
Revises: 48dd6f2e0784
Create Date: 2014-03-14 15:29:22.379106

"""

# revision identifiers, used by Alembic.
revision = '3a3c75de71b'
down_revision = '48dd6f2e0784'

from alembic import op
import sqlalchemy as sa


def upgrade():
    op.add_column('grano_attribute', sa.Column('datatype', sa.Unicode(), nullable=True))
    op.alter_column('grano_property', 'value', new_column_name='value_string')
    
    attribute = sa.sql.table('grano_attribute',
        sa.sql.column('datatype', sa.Unicode)
    )

    op.execute(attribute.update().values({'datatype': 'string'}))


    # leftovers:
    op.create_index('ix_grano_property_entity_id', 'grano_property', ['entity_id'], unique=False)
    op.create_index('ix_grano_property_name', 'grano_property', ['name'], unique=False)
    op.create_index('ix_grano_property_relation_id', 'grano_property', ['relation_id'], unique=False)
    op.drop_index('ix_property_entity_id', table_name='grano_property')
    op.drop_index('ix_property_name', table_name='grano_property')
    op.drop_index('ix_property_relation_id', table_name='grano_property')
    op.create_index('ix_grano_relation_schema_id', 'grano_relation', ['schema_id'], unique=False)
    op.create_index('ix_grano_relation_source_id', 'grano_relation', ['source_id'], unique=False)
    op.create_index('ix_grano_relation_target_id', 'grano_relation', ['target_id'], unique=False)
    op.drop_index('ix_relation_schema_id', table_name='grano_relation')
    op.drop_index('ix_relation_source_id', table_name='grano_relation')
    op.drop_index('ix_relation_target_id', table_name='grano_relation')


def downgrade():
    op.create_index('ix_relation_target_id', 'grano_relation', ['target_id'], unique=False)
    op.create_index('ix_relation_source_id', 'grano_relation', ['source_id'], unique=False)
    op.create_index('ix_relation_schema_id', 'grano_relation', ['schema_id'], unique=False)
    op.drop_index('ix_grano_relation_target_id', table_name='grano_relation')
    op.drop_index('ix_grano_relation_source_id', table_name='grano_relation')
    op.drop_index('ix_grano_relation_schema_id', table_name='grano_relation')
    op.create_index('ix_property_relation_id', 'grano_property', ['relation_id'], unique=False)
    op.create_index('ix_property_name', 'grano_property', ['name'], unique=False)
    op.create_index('ix_property_entity_id', 'grano_property', ['entity_id'], unique=False)
    op.drop_index('ix_grano_property_relation_id', table_name='grano_property')
    op.drop_index('ix_grano_property_name', table_name='grano_property')
    op.drop_index('ix_grano_property_entity_id', table_name='grano_property')
    op.drop_column('grano_attribute', 'datatype')

    op.alter_column('grano_property', 'value_string', new_column_name='value')




"""empty message

Revision ID: d7067cd4169
Revises: 5a34d3b6a160
Create Date: 2014-03-14 16:25:25.820714

"""

# revision identifiers, used by Alembic.
revision = 'd7067cd4169'
down_revision = '5a34d3b6a160'

from alembic import op
import sqlalchemy as sa


def upgrade():
    op.add_column('grano_property', sa.Column('attribute_id', sa.Integer(), nullable=True))

    attributes = sa.sql.table('grano_attribute',
        sa.sql.column('id', sa.Integer),
        sa.sql.column('schema_id', sa.Integer),
        sa.sql.column('name', sa.Unicode)
    )

    properties = sa.sql.table('grano_property',
        sa.sql.column('schema_id', sa.Integer),
        sa.sql.column('attribute_id', sa.Integer),
        sa.sql.column('name', sa.Unicode)
    )

    connection = op.get_bind()
    rp = connection.execute(attributes.select())

    if rp is None:
        return None

    rows = [dict(zip(r.keys(), tuple(r))) for r in rp.fetchall()]
    for row in rows:
        #print row
        q = properties.update()
        q = q.where(properties.c.schema_id==row.get('schema_id'))
        q = q.where(properties.c.name==row.get('name'))
        q = q.values({'attribute_id': row.get('id')})
        #print q, row.get('id')
        op.execute(q)

    q = properties.delete()
    q = q.where(properties.c.attribute_id==None)
    op.execute(q)


def downgrade():
    ### commands auto generated by Alembic - please adjust! ###
    op.drop_column('grano_property', 'attribute_id')
    ### end Alembic commands ###



import json
from datetime import datetime
from uuid import UUID
from decimal import Decimal
from inspect import isgenerator

from flask import Response, request


class JSONEncoder(json.JSONEncoder):
    """ This encoder will serialize all entities that have a to_dict
    method by calling that method and serializing the result. """

    def __init__(self, index=False):
        self.index = index
        super(JSONEncoder, self).__init__()

    def default(self, obj):
        if self.index and hasattr(obj, 'to_dict_index'):
            return obj.to_dict_index()
        if hasattr(obj, 'to_dict'):
            return obj.to_dict()
        if isinstance(obj, datetime):
            return obj.isoformat() + 'Z'
        if isinstance(obj, UUID):
            return str(obj)
        if isinstance(obj, Decimal):
            return float(obj)
        if isgenerator(obj):
            return [o for o in obj]
        return json.JSONEncoder.default(self, obj)


def jsonify(obj, status=200, headers=None, index=False, encoder=JSONEncoder):
    """ Custom JSONificaton to support obj.to_dict protocol. """
    if encoder is JSONEncoder:
        data = encoder(index=index).encode(obj)
    else:
        data = encoder().encode(obj)
    if 'callback' in request.args:
        cb = request.args.get('callback')
        data = '%s && %s(%s)' % (cb, cb, data)
    return Response(data, headers=headers,
                    status=status,
                    mimetype='application/json')



from grano.core import db
from grano.manage import app
from grano.test.fixtures import create_fixtures


def make_test_app(use_cookies=False):
    app.config['TESTING'] = True
    app.config['CACHE'] = False
    app.config['PLUGINS'] = []
    app.config['CELERY_ALWAYS_EAGER'] = True
    app.config['SQLALCHEMY_DATABASE_URI'] = 'sqlite:///:memory:'
    db.create_all()

    return app.test_client(use_cookies=use_cookies)


def tear_down_test_app():
    #db.session.rollback()
    db.drop_all()




import colander
from datetime import datetime

from grano.core import app, db, celery
from grano.logic.validation import database_name
from grano.logic.references import AccountRef
from grano.plugins import notify_plugins
from grano.model import Project


def validate(data, project):
    same_project = lambda s: Project.by_slug(s) == project
    same_project = colander.Function(same_project, message="Project exists")

    class ProjectValidator(colander.MappingSchema):
        slug = colander.SchemaNode(colander.String(),
                                   validator=colander.All(database_name,
                                                          same_project))
        label = colander.SchemaNode(colander.String(),
                                    validator=colander.Length(min=3))
        private = colander.SchemaNode(colander.Boolean(),
                                      missing=False)
        author = colander.SchemaNode(AccountRef())
        settings = colander.SchemaNode(colander.Mapping(),
                                       missing={})

    validator = ProjectValidator()
    return validator.deserialize(data)


@celery.task
def _project_changed(project_slug, operation):
    """ Notify plugins about changes to a relation. """
    def _handle(obj):
        obj.project_changed(project_slug, operation)
    notify_plugins('grano.project.change', _handle)


def save(data, project=None):
    """ Create or update a project with a given slug. """

    data = validate(data, project)

    operation = 'create' if project is None else 'update'
    if project is None:
        project = Project()
        project.slug = data.get('slug')
        project.author = data.get('author')

        from grano.logic import permissions as permissions_logic
        permissions_logic.save({
            'account': data.get('author'),
            'project': project,
            'admin': True
            })

    project.settings = data.get('settings')
    project.label = data.get('label')
    project.private = data.get('private')
    project.updated_at = datetime.utcnow()

    db.session.add(project)

    # TODO: make this nicer - separate files?
    from grano.logic.schemata import import_schema
    with app.open_resource('fixtures/base.yaml') as fh:
        import_schema(project, fh)

    db.session.flush()
    _project_changed(project.slug, operation)
    return project


def delete(project):
    """ Delete the project and all related data. """
    _project_changed(project.slug, 'delete')
    db.session.delete(project)


def truncate(project):
    """ Delete all entities and relations from this project,
    but leave the project, schemata and attributes intact. """
    from grano.logic import relations
    from grano.logic import entities

    project.updated_at = datetime.utcnow()

    for relation in project.relations:
        relations.delete(relation)

    for entity in project.entities:
        entities.delete(entity)
